#include "typedef.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "util.h"
#include "object.h"
#include "class.h"
#include "method.h"

extern int g_subclass_permissions;

Cmethod	*Cmethod::g_hash[HASHSIZE] = {0};

/* static */ u4T
Cmethod::hash(const u2T *nameP, const u2T *descriptorP)
{
	int		ret  = 0;
	u2T		c;
	const	u2T *P;

	for (P = nameP; c = *P ; ++P) {
		ret  += c;
		ret <<= 1;
		if (ret & HASHSIZE) {
			ret |= 1;
	}	}
	for (P = descriptorP; c = *P ; ++P) {
		ret  += c;
		ret <<= 1;
		if (ret & HASHSIZE) {
			ret |= 1;
	}	}
	return(ret & (HASHSIZE-1));
}

Cmethod::Cmethod(Cclass *classP, const u2T *nameP, const u2T *descriptorP)
{
	m_hashP        = 0;
	m_parentP      = classP;
	m_nameP        = unicodedup(nameP);
	m_descriptorP  = unicodedup(descriptorP);
	m_access_flags = 0;
	m_max_stack    = 0;
	m_max_locals   = 0;
	m_lineno       = -1;
	m_throws_cnt   = 0;
	m_throwsPP     = 0;
}

Cmethod *
Cmethod::locate(Cclass *classP, const u2T *nameP, const u2T *descriptorP)
{
	u4T			index;
	Cmethod	**methodPP, *methodP;

	index = hash(nameP, descriptorP);
	for (methodPP = g_hash + index; methodP = *methodPP; methodPP = &(methodP->m_hashP)) {
		if (classP == methodP->m_parentP && !unicodecmp(methodP->m_nameP, nameP) && !unicodecmp(methodP->m_descriptorP, descriptorP) ) {
			return(methodP);
	}	}
	methodP = new Cmethod(classP, nameP, descriptorP);
	if (!methodP) {
		outofmemory();
	}
	*methodPP = methodP;
	return(methodP);
}

/* virtual */ objectE
Cmethod::type(void) const
{
	return(functionE);
}

/* virtual */ void
Cmethod::emit_id(void) const
{
	fprintf(stdout, "F%p", this);
}		

/* virtual */ void
Cmethod::emit_edges(void)
{
	int	throws_cnt;

	Centity::emit_edges();

	if ((throws_cnt = m_throws_cnt)) {
		Cclass	**throwsPP = m_throwsPP;
		Cclass	*throwsP;
		int		i;

		for (i = 0; i < throws_cnt; ++i) {
			throwsP = throwsPP[i];
			fputs("E261 ", stdout);
			emit_id();
			fputc(' ', stdout);
			throwsP->emit_id();
			fputc('\n', stdout);
}	}	}		

const char *
Cmethod::get_type(void) const
{
	const char	*typeP;

	if (g_subclass_permissions) {
		u2T		access = m_access_flags;

		if (access & ACC_PRIVATE) {
			if (access & ACC_PROTECTED) {
				typeP = "M6";
			} else {
				typeP = "M2";
			}
		} else if (access & ACC_PROTECTED) {
			typeP = "M4";
		} else if (access & ACC_PUBLIC) {
			typeP = "M1";
		} else {
			typeP = "M0";
		}
	} else {
		typeP = "M";
	}
	return typeP;
}

/* virtual */ void
Cmethod::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE F%p %s\n", this, get_type());
}

void
Cmethod::emit_attributes(void) const
{
	emit_id();

	fputs(" {label=\"", stdout);
	dumpUnicode(m_nameP);
	fputc('"', stdout);
	fputs(" file=.", stdout);
	if (m_lineno >= 0) {
		fprintf(stdout, " lineno=%d", m_lineno);
	}

	if (m_descriptorP) {
		fputs(" descriptor=\"", stdout);
		dump_descriptor(m_descriptorP);
		fputc('"', stdout);
	}
	dump_access(m_access_flags);

	if (m_max_stack) {
		fprintf(stdout, " maxstack=%u", m_max_stack);
	}
	if (m_max_locals) {
		fprintf(stdout, " maxlocals=%u", m_max_locals);
	}
	fputs("}\n", stdout);
}
